using System;
using System.Collections.Generic;
using System.Threading;
using SlatedGameToolkit.Framework.StateSystem.States;

namespace SlatedGameToolkit.Framework.StateSystem
{
    public sealed class Manager {
        public Thread thread;
        private IState currentState;
        private IState nextState;
        private Dictionary<string, IState> states;

        /// <summary>
        /// Instantiates a game state manager with an initial state, and a set of states to be added at the start.
        /// States can later be added or removed.
        /// None of the parameters can be null, and the initial state must exist in initial set of states.
        /// </summary>
        /// <param name="initialState">The name of the initial state.</param>
        /// <param name="states">The initial set of game states to be added.</param>
        public Manager(string initialState, params IState[] states) {
            if (initialState == null) throw new ArgumentNullException("initialState");
            this.states = new Dictionary<string, IState>();
            thread = Thread.CurrentThread;
            addStates(states);
            if (!this.states.TryGetValue(initialState, out currentState)) throw new ArgumentException("The requested initial state name does not exist in the provided list of states.");
        }

        internal void initialize() {
            thread = Thread.CurrentThread;
        }

        internal void update(double delta) {
            if (nextState != null) {
                if (currentState.Deactivate() & nextState.Activate()) {
                    currentState = nextState;
                    nextState = null;
                }
            }
            currentState.Update(delta);
        }

        internal void render(double delta) {
            currentState.Render(delta);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="name"></param>
        public void changeState(string name) {
            if (thread != Thread.CurrentThread) throw new ThreadStateException("State cannot be changed from a different thread.");
            if (!states.TryGetValue(name, out nextState)) throw new ArgumentException("The requested state to change to does not exist in this manager.");
        }

        /// <summary>
        /// Adds the given states to this manager under each of their own respective names.
        /// Initializes all the states.
        /// If there are two states of the same name, the first state in the array is added and the second is discarded.
        /// The discarded state will not be initialized and dispose will not be called on it.
        /// </summary>
        /// <param name="states">The states to add.</param>
        /// <returns>True if all the states were unique, and false if there were repetitions determined by name.</returns>
        public bool addStates(IState[] states) {
            if (states == null || states.Length == 0) throw new ArgumentException("The array of states cannot be null, and cannot be of size 0");
            bool unique = true;
            foreach (IState state in states)
            {
                unique = unique && addState(state) ? true : false;
            }
            return unique;
        }

        /// <summary>
        /// Adds the given state to this manager under it's own name.
        /// Initializes the state as well.
        /// Will not iniitialize the state if this method returns false.
        /// </summary>
        /// <param name="state">The state to be added to this manager.</param>
        /// <returns>False if a state of this name has already been registered.</returns>
        public bool addState(IState state) {
            if (thread != Thread.CurrentThread) throw new ThreadStateException("Cannot add a state from a different thread.");
            try {
            this.states.Add(state.getName(), state);
            } catch (ArgumentException) {
                return false;
            }
            state.Initialize(this);
            return true;
        }

        /// <summary>
        /// Removes the state given the name of the state.
        /// If the state doesn't exist, nothing happens.
        /// Will call dispose on the state.
        /// </summary>
        /// <param name="name">The name of the state.</param>
        /// <returns>False if the state is being used, or the state doesn't exist.</returns>
        public bool removeState(string name) {
            if (thread != Thread.CurrentThread) throw new ThreadStateException("Cannot remove a state from a different thread.");
            if (states[name] == currentState) return false;
            IState state = states[name];
            state.Dispose();
            return states.Remove(name);
        }

        /// <summary>
        /// Removes all the states in this state manager except for the current one.
        /// Disposes of the removed states.
        /// </summary>
        public void removeAllStates() {
            foreach (String state in this.states.Keys)
            {
                removeState(state);
            }
        }
    }
}